#!/bin/sh
# Tcl magic \
exec wish8.2 -f $0 $*
################################################################################
#
# Copyright (C) 1996, 1997
#      Michael Smith.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY Michael Smith AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL Michael Smith OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
################################################################################
#
# Ports Index Browser
#
################################################################################
#
# A tool for browsing the FreeBSD ports INDEX file.
#
# Lines in the index are formatted as follows :
#
# <name>|<path>|<prefix>|<comment>|<descriptionpath>|<maintainer>|<category>|<build-depends>|<run-depends>
#
# When parsing the file, we build two arrays :
#
# Paths() contains elements which are the paths containing ports, and each
# element's value is the name of the port at that path.
#
# Ports() contains elements whose names, given the port is called <port> are :
#
# <port>:name		the port name ("<port>")
# <port>:path		path of the port
# <port>:prefix		where the port installs
# <port>:comment	brief description
# <port>:dpath		path to DESCR file
# <port>:maint		maintainer
# <port>:cat		category
# <port>:bdep		build-depends
# <port>:rdep		run-depends
#
################################################################################
#
# Arrays used for holding distfile information :
#
# AllDistfiles() contains elements which are the paths of all possible 
# distfiles (required and present), and whose values are lists of the port(s) 
# that use the distfiles.
#
# DistfileStatus() contains the same set of elements, and the value of the
# element indicates the status of the distfile (stale etc.)
#
################################################################################
#
# If an alternate filename is given on the commandline, open it instead of
# /usr/ports/INDEX
#
proc main {} {

    global argc argv argv0 Options Status;

    # Check up on ourselves
    if {[catch {set ver [info tclversion]}]} {
	set ver 0;
    }
    if {$ver < 7.5} {
	puts "Sorry, needs Tcl 7.5 or later";
	exit;
    }	
    # Find our base directory
    set pname $argv0;
    while {[file type $pname] == "link"} {
	set pname [file readlink $pname];
    }
    set Status(homedir) [file dirname $pname];

    # Load the tkSteal extension
    load [glob $Status(homedir)/libTkSteal.so.2*];
    source $Status(homedir)/tkSteal.tcl;

    # initialise Options
    op_init;

    # setup our help
    h_init;

    # find our INDEX
    set ipath "[op_elookup PORTSDIR]/INDEX";
    if {$argc == 1} {
	set ipath [lindex $argv 0];
    }

    if {![file readable $ipath]} {
	puts "Can't read INDEX file '$ipath'";
	exit ;
    }
    if {[catch {set fh [open $ipath "r"]} msg]} {
	puts $msg;
	exit ;
    }

    # Build the browser window, so the default toplevel isn't
    # seen, because it looks stupid.
    wm title "." "Ports INDEX browser";
    init_options;
    create_browser "";

    # put up the progress window
    create_pwin "iparse" "Reading INDEX, %s ports...";

    # parse the INDEX
    read_index $fh "iparse";
    close $fh;
    
    # nuke the progress-counter
    destroy_pwin "iparse"

    # update the browser, let it run.
    init_browser;
}

################################################################################
# create_browser
#
# Build the toplevel browser inside (w)
#
proc create_browser {w} {

    global Browser;

    # Add the menu
    frame .menuBar -relief raised -bd 2;
    pack .menuBar -side top -fill x;

    # And things on it
    menubutton .menuBar.func -text "Function" -menu .menuBar.func.m -underline 0;
    menu .menuBar.func.m -tearoff 0;
    .menuBar.func.m add command -label "Search..." -command "br_search" -underline 0;
    .menuBar.func.m add command -label "Distfiles..." -command "df_browser" -underline 0;
    .menuBar.func.m add command -label "Maker..." -command "mk_create" -underline 0;
    .menuBar.func.m add command -label "Options..." -command "op_set" -underline 0;
    .menuBar.func.m add separator;
    .menuBar.func.m add command -label "Quit" -command exit -underline 0;
    pack .menuBar.func -side left;

    menubutton .menuBar.help -text "Help" -menu .menuBar.help.m -underline 0;
    menu .menuBar.help.m -tearoff 0;
    .menuBar.help.m add command -label "Using..." -command "help browser" -underline 0;
    .menuBar.help.m add command -label "About..." -command "help about" -underline 0;
    pack .menuBar.help -side right;

    # Path frame
    frame $w.path;
    pack $w.path -side top -fill x;
    set sf $w.path;
    label $sf.l -text "Path : ";
    label $sf.p -textvariable Browser(cpath) -relief sunken -bd 2 -anchor w \
	-background white;
    button $sf.b -text "<-" -command "br_back" -padx 3 -pady 0;
    pack $sf.l -side left;
    pack $sf.p -side left -fill x -expand y;
    pack $sf.b -side right;

    # Main console frame.
    frame $w.cons;
    pack $w.cons -side top -fill both -expand y;
    
    # path/port selector frame
    frame $w.cons.sel;
    pack $w.cons.sel -side left -fill y;
    set sf $w.cons.sel;

    listbox $sf.list -width 30 -yscrollcommand "$sf.scroll set";
    scrollbar $sf.scroll -command "$sf.list yview";
    bind $sf.list <Double-1> "br_forward";
    pack $sf.list $sf.scroll -side left -fill y;
    set Browser(sellist) $sf.list;

    # port info frame
    frame $w.cons.port;
    pack $w.cons.port -side left -fill both -expand y;
    set sf $w.cons.port;

    # name, 'get description' function
    frame $sf.name;
    pack $sf.name -side top -fill x;
    label $sf.name.l -textvariable Browser(pname) -relief raised -width 35 -bd 1;
    pack $sf.name.l -side left -fill x -expand y;
    button $sf.name.b -text "?" -command "br_desc new" -padx 2 -pady 0;
    pack $sf.name.b -side right;
    
    # comment
    text $sf.desc -wrap word -height 2 -relief flat -bd 0 -state disabled -width 50;
    pack $sf.desc -side top -fill both;
    set Browser(pcomment) $sf.desc;

    # maintainer
    frame $sf.maint -relief groove -bd 1;
    pack $sf.maint -side top -fill x;
    label $sf.maint.l -text "Maintainer : ";
    label $sf.maint.a -textvariable Browser(maint);
    pack $sf.maint.l $sf.maint.a -side left;

    # current status
    frame $sf.stat -relief groove -bd 1;
    pack $sf.stat -side top -fill x;
    set ssf $sf.stat;

    label $ssf.absent -text "absent" -foreground gray;
    label $ssf.clean -text "clean" -foreground gray;
    label $ssf.extract -text "extract" -foreground gray;
    label $ssf.patch -text "patch" -foreground gray;
    label $ssf.configure -text "configure" -foreground gray;
    label $ssf.build -text "build" -foreground gray;
    label $ssf.install -text "install" -foreground gray;
    label $ssf.package -text "package" -foreground gray;

    pack $ssf.absent $ssf.clean $ssf.extract $ssf.configure $ssf.build \
	$ssf.install $ssf.package -side left;

    # save all these paths where others can get them
    set Browser(stat:absent)	$ssf.absent;
    set Browser(stat:clean)	$ssf.clean;
    set Browser(stat:extract)	$ssf.extract;
    set Browser(stat:patch)	$ssf.patch;
    set Browser(stat:configure) $ssf.configure;
    set Browser(stat:build)	$ssf.build;
    set Browser(stat:install)	$ssf.install;
    set Browser(stat:package)	$ssf.package;

    # Depend lists
    frame $sf.dep;
    pack $sf.dep -side top -fill both -expand y;
    
    # build depends
    frame $sf.dep.b -relief groove -bd 1;
    pack $sf.dep.b -side left -fill both -expand y;
    label $sf.dep.b.t -text "Build requires" -anchor w;
    pack $sf.dep.b.t -side top -fill x;
    frame $sf.dep.b.l;
    pack $sf.dep.b.l -side top -fill both -expand y;
    set ssf $sf.dep.b.l;

    listbox $ssf.list -width 15 -height 5 -yscrollcommand "$ssf.scroll set";
    scrollbar $ssf.scroll -command "$ssf.list yview";
    bind $ssf.list <Double-1> "br_xref $ssf.list";
    pack $ssf.list -side left -fill both -expand y;
    pack $ssf.scroll -side left -fill y;
    set Browser(bdlist) $ssf.list;

    # run depends
    frame $sf.dep.r -relief groove -bd 1;
    pack $sf.dep.r -side left -fill both -expand y;
    label $sf.dep.r.t -text "Run requires" -anchor w;
    pack $sf.dep.r.t -side top -fill x;
    frame $sf.dep.r.l;
    pack $sf.dep.r.l -side top -fill both -expand y;
    set ssf $sf.dep.r.l;

    listbox $ssf.list -width 15 -height 5 -yscrollcommand "$ssf.scroll set";
    scrollbar $ssf.scroll -command "$ssf.list yview";
    bind $ssf.list <Double-1> "br_xref $ssf.list";
    pack $ssf.list -side left -fill both -expand y;
    pack $ssf.scroll -side left -fill y;
    set Browser(rdlist) $ssf.list;
}

################################################################################
# init_browser
#
# Set the initial browser etc. state.
#
proc init_browser {} {

    global Browser Options Status;

    # Browser state
    set Browser(cpath) "[op_elookup PORTSDIR]/";	# initial path
    set Browser(opath) "";				# no previous path
    set Browser(history) "";				# no back-traversal history
    br_selset;						# set the selector up
    br_setport "";

    # distfile state
    set Status(df:info) "invalid";
}

################################################################################
# init_options
#
# set options for our look-and-feel
#
proc init_options {} {

    # scrollbars
    option add *Scrollbar.width 11;

    # textboxes
    option add *Text.font fixed;
    
}

################################################################################
# Help texts
#
proc h_init {} {

#               ###########################################################################
    help_init "browser" \
	[list \
	     "INDEX Browser" \
	     "=============" \
	     "" \
	     "  The INDEX browser provides a quick and powerful means for browsing the" \
	     "  ports INDEX file and performing related functions.  Navigate the ports " \
	     "  heirachy by double-clicking the browse list on the left.  The Path field "\
	     "  always displays your current location." \
	     "" \
	     "  Items in the browse list with a trailing slash are directories, those" \
	     "  without are ports.  The '..' entry will return one level up the heirachy." \
	     "" \
	     "  Selecting a port will show its details on the right.  The category and" \
	     "  title are displayed at the top, and clicking the \[?\] button will open" \
	     "  a window containing a more verbose description of the port." \
	     "" \
	     "  Entries in the 'Build requires' and 'Run requires' lists are 'hot'," \
	     "  double-clicking one will jump to that port.  Clicking the \[<-\] button" \
	     "  will unwind through cross-references like this, and then move back up" \
	     "  the heirachy." \
	     "" \
	     "  Additional windows (help, description etc.) are persistent; once open" \
	     "  they will update to reflect new ports as they are selected." \
	    ];

#               ###########################################################################
    help_init "about" \
	[list \
	     "" \
	     "" \
	     "                     Ports INDEX Browser" \
	     "                     ===================" \
	     "" \
	     "             Copyright (c) 1996,1997 Michael Smith" \
	     "                     msmith@FreeBSD.org" \
	     "" \
	     "           Please report any bugs to the above address." \
	     "            The newest version is available from your nearest"\
	     "                      FreeBSD ports mirror."\
	     "" \
	     "  Redistribution and use in source and binary forms, with or without" \
	     "  modification, are permitted provided that the following conditions" \
	     "  are met:" \
	     "  1. Redistributions of source code must retain the above copyright" \
	     "     notice, this list of conditions and the following disclaimer." \
	     "  2. Redistributions in binary form must reproduce the above copyright" \
	     "     notice, this list of conditions and the following disclaimer in the" \
	     "     documentation and/or other materials provided with the distribution." \
	     "" \
	     "  THIS SOFTWARE IS PROVIDED BY Michael Smith AND CONTRIBUTORS ``AS IS'' AND" \
	     "  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE" \
	     "  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE" \
	     "  ARE DISCLAIMED.  IN NO EVENT SHALL Michael Smith OR CONTRIBUTORS BE LIABLE" \
	     "  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL" \
	     "  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS" \
	     "  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)" \
	     "  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT" \
	     "  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY" \
	     "  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF" \
	     "  SUCH DAMAGE." \
	     "" \
	     { $Id: pib.tcl,v 1.14 1997/09/11 09:16:56 msmith Exp $ } \
	    ];

#               ###########################################################################
    help_init "search" \
	[list \
	     "Searching" \
	     "=========" \
	     "" \
	     "  The INDEX can be searched on any one or more of the selectable fields, and" \
	     "  the results directly jumped to." \
	     "" \
	     "  Search terms are substrings, and may contain wildcards in csh(1) 'glob'" \
	     "  syntax : '?' matches any single character, '*' matches any characters," \
	     "  \[a-z\] matches any character between 'a' and 'z' inclusive." \
	     "" \
	     "  Search results are listed in the browser.  These results are 'hot', so" \
	     "  double-clicking on a port name will jump to that port.  The search dialog" \
	     "  stays open until explicitly closed to allow easy browsing of all matches" \
	     "  to a search." \
	    ];

#               ###########################################################################
    help_init "options" \
	[list \
	     "Options" \
	     "=======" \
	     "" \
	     "  Options controlling the behaviour of the ports tree in general, and the " \
	     "  browser in particular can be set here.  These values are read from the" \
	     "  file ~/.pib at startup, and can be saved with the \[Save\] button." \
	     "" \
	     "  Port variable defaults" \
	     "  ----------------------" \
	     "  Default values for environment variables used by the port system can be" \
	     "  set here.  If these values are present in the environment or in " \
	     "  /etc/make.conf, the values here are overridden." \
	     "  Consult /usr/share/mk.bsd.port.mk for more details on these options." \
	     "" \
	     "  Port structure warnings" \
	     "  -----------------------" \
	     "  These options control the output of warning messages when processing" \
	     "  the ports tree structure.  Warnings are displayed in a browser at the" \
	     "  completion of processing, which is performed the first time the " \
	     "  Distfile browser is opened, or when the \[Rescan\] function is selected." \
	     "" \
	     "  No MD5 file            The port's 'files' directory has no 'MD5' file." \
	     "  No files directory     The port has no 'files' directory" \
	     "  Port directory missing The port's skeleton directory is missing" \
	     "" \
	     "  Runtime warnings" \
	     "  ----------------" \
	     "  Some runtime conditions can generate error messages, which can be" \
	     "  selectively suppressed here." \
	     "" \
	     "  Duplicate port names   Ports normally may not share names.  Any subsequent" \
	     "                         ports after the first using a name are ignored." \
	     "  Errors deleting files  When purging stale distfiles, errors may be " \
	     "                         encountered deleting files." \
	     "" \
	     "  Action options" \
	     "  --------------" \
	     "  Some actions may be undesirable, and can be disabled here." \
	     "" \
	     "  Really delete files    If this is not set, a report of stale distfiles that" \
	     "                         would have been deleted is produced instead of" \
	     "                         actually deleting the files." \
	     ];

#               ###########################################################################
    help_init "distfiles" \
	[list \
	     "Distfile browser" \
	     "================" \
	     "" \
	     "  Distfiles are grouped in three categories :" \
	     "" \
	     "  Current       Present in the distfile directory and up to date." \
	     "  Stale         Present, but not required by any port, or with an invalid" \
	     "                checksum, implying that the file has changed on the master" \
	     "                site but not the name of the file." \
	     "  Missing       Not present, but required for one or more ports." \
	     "" \
	     "  Note that a distfile that is stale due to a bad checksum is not considered" \
	     "  to be missing.  The correctness of this is debatable." \
	     "" \
	     "  Any combination of categories can be selected with the controls at the " \
	     "  top of the display.  The 'For selected port' option restricts the display" \
	     "  further to just those distfiles that are relevant to the port currently" \
	     "  displayed in the INDEX browser.  This button also constrains the action" \
	     "  of the \[Delete\] button to stale files associated with the current port." \
	     "" \
	     "  Distfiles are listed colour-coded for category on the left, and extra " \
	     "  status on the right.  The checkbox at the beginning of the status entry " \
	     "  is marked if the file is tagged, clicking on a checkbox toggles the tag " \
	     "  state for a distfile.  Only stale distfiles can be tagged." \
	     "" \
	     "  Stale distfiles have a reason shown in brackets, eg. (stale) or (bad MD5)." \
	     "  If a distfile is required by one or more ports, these are listed in the" \
	     "  status area." \
	     "" \
	     "  \[Dump\] generates a plain-text listing of the currently-visible distfiles" \
	     "  and their status in the file 'distfiles.dump in the current directory." \
	     "" \
	     "  \[Rescan\] re-scans the distfiles directory and the ports collection to " \
	     "  incorporate any changes that may have occurred." \
	     "" \
	     "  \[Delete\] deletes any tagged stale distfiles, subject to the setting of" \
	     "  the 'For selected port' control.  If the 'Really delete files' option is" \
	     "  not enabled, a list of files that would have been deleted is produced" \
	     "  instead." \
	     "" \
	     "  \[Tag\]/\[Untag\] sets/clears the tagged state of all visible distfiles." \
	    ];

#               ###########################################################################
    help_init "maker" \
	[list \
	     "Port maker" \
	     "==========" \
	     "" \
	     "  The ultimate goal of the Ports Collection is to build and install software," \
	     "  and this is what the Maker is for.  Central to the window is a captive" \
	     "  xterm, in which all building operations are performed." \
	     "" \
	     "  At top left is the status indicator, showing what the Maker is currently" \
	     "  doing.  At top right is the name of the currently-selected port, which" \
	     "  will be operated on by the function buttons listed below it." \
	     "" \
	     "  The standard functions are :" \
	     "" \
	     "  Clean     Remove all working files." \
	     "  Fetch     Download the distribution files and any patches." \
	     "  Extract   Unpack the distribution files." \
	     "  Patch     Apply any necessary patches." \
	     "  Configure Perform any configuration required before building." \
	     "  Build     Compile the software." \
	     "  Install   Install the software (must be root)." \
	     "  Package   Make a package file from the installed software." \
	     "" \
	     "  Once a function has been selected, the function buttons will be disabled" \
	     "  and the \[Accept\] button enabled.  Clicking the \[Accept\] button" \
	     "  reenables the function buttons, and this should be performed once the" \
	     "  user is satisfied that the function is complete." \
	     "" \
	     "  In addition, commands can be typed directly into the xterm as normal." \
	     "  Avoid logging out in the xterm; the Maker window will have to be" \
	     "  closed and reopened in order to obtain a new xterm in this case." \
	    ];
}

################################################################################
# Options
################################################################################
# op_set
#
# Options control dialog
#
proc op_set {} {

    global Options Envar;

    set w .options;

    if {[winfo exists $w]} {			# already exists
	wm deiconify $w;			# do our best to become visible
	raise $w;
	tkwait visibility $w;
	focus $w;
	return ;
    }

    catch {destroy $w};
    toplevel $w;
    wm title $w "PIB Options";

    label $w.t1 -text "Port variable defaults" -relief raised -bd 1;
    pack $w.t1 -side top -fill x;

    foreach tag $Envar(all) {
	op_evar $w $tag;
    }

    label $w.t2 -text "Port structure warnings" -relief raised -bd 1;
    pack $w.t2 -side top -fill x;

    # Options
    set sf [op_sframe $w "w_md5" "No MD5 file"];
    checkbutton $sf.w -variable Options(warn_no_md5);
    pack $sf.w -side left;

    set sf [op_sframe $w "w_files" "No files directory"];
    checkbutton $sf.w -variable Options(warn_no_fdir);
    pack $sf.w -side left;

    set sf [op_sframe $w "w_pres" "Port directory missing"];
    checkbutton $sf.w -variable Options(warn_no_port);
    pack $sf.w -side left;

    label $w.t3 -text "Runtime warnings" -relief raised -bd 1;
    pack $w.t3 -side top -fill x;

    set sf [op_sframe $w "w_dupnames" "Duplicate port names"];
    checkbutton $sf.w -variable Options(warn_dup_names);
    pack $sf.w -side left;

    set sf [op_sframe $w "w_del" "Errors deleting files"];
    checkbutton $sf.w -variable Options(warn_del_errs);
    pack $sf.w -side left;

    label $w.t4 -text "Action options" -relief raised -bd 1;
    pack $w.t4 -side top -fill x;

    set sf [op_sframe $w "d_del" "Really delete files"];
    checkbutton $sf.w -variable Options(do_deletions);
    pack $sf.w -side left;

    # buttons
    frame $w.b;
    pack $w.b -side bottom -fill x -expand y;

    button $w.b.save -text "Save" -width 10 -command "op_save";
    button $w.b.hlp -text "Help" -command "help options";
    button $w.b.dis -text "Dismiss" -width 10 -command "destroy $w";
    pack $w.b.save -side left;
    pack $w.b.dis $w.b.hlp -side right;
}

################################################################################
# op_sframe
#
# Given (w), pack a subframe in it identified by (tag).  Pack a label (text) to
# the left, and return the path of the subframe.
#
proc op_sframe {w tag text} {

    frame $w.$tag;
    pack $w.$tag -side top -fill x;
    label $w.$tag.l -text $text -width 25 -anchor e;
    pack $w.$tag.l -side left;
    return $w.$tag;
}

################################################################################
# op_evar
#
# Given (w), pack into it an editor for the environment variable indexed
# with (name)
#
proc op_evar {w name} {

    global Options Envar;

    # Label and subframe
    set sf [op_sframe $w $name $Envar($name:name)];
    
    if {[lsearch $Envar($name:allow) "unset"] != -1} {
	# pull the focus off any entryfields after this
	radiobutton $sf.unset -text "Unset" -variable Envar($name:state) -value "unset" \
	    -width 5 -anchor w -command "set Envar($name:value) {}; focus $sf";
	pack $sf.unset -side left;
    }
    if {[lsearch $Envar($name:allow) "set"] != -1} {
	radiobutton $sf.set -text "Set" -variable Envar($name:state) -value "set" \
	    -width 5 -anchor w;
	pack $sf.set -side left;
    }
    if {[lsearch $Envar($name:allow) "yesno"] != -1} {
	radiobutton $sf.no -text "NO" -variable Envar($name:state) -value "no" \
	    -width 5 -anchor w;
	radiobutton $sf.yes -text "YES" -variable Envar($name:state) -value "yes" \
	    -width 5 -anchor w;
	pack $sf.no $sf.yes  -side left;
    }
    if {[lsearch $Envar($name:allow) "value"] != -1} {
	# send the focus to the entryfield for instant input
	radiobutton $sf.ven -text "Value" -variable Envar($name:state) -value "value" \
	    -width 5 -anchor w -command "focus $sf.val";
	entry $sf.val -textvariable Envar($name:value) -width 30;
	pack $sf.ven -side left;
	pack $sf.val -side left -fill x -expand y;
    }
}

################################################################################
# op_edef
#
# Set up definitions for environment variables
#
proc op_edef {tag name allow {defstate unset} {defval ""}} {

    global Envar;

    set Envar($tag:name) $name;
    set Envar($tag:allow) $allow;
    set Envar($tag:state) $defstate;
    set Envar($tag:value) $defval;

    lappend Envar(all) $tag;
}

################################################################################
# op_estatus
#
# Given the name of an environment variable (name), update its status to 
# match its value.  Report unhappiness using (pfx) as a prefix.
#
proc op_estatus {name pfx} {

    global Envar;

    set Envar($name:state) "unknown";
    if {[lsearch $Envar($name:allow) set] != -1} {
	set Envar($name:state) set;
    }
    if {[lsearch $Envar($name:allow) yesno] != -1} {
	if {$Envar($name:value) == "YES"} {
	    set Envar($name:state) yes;
	}
	if {$Envar($name:value) == "NO"} {
	    set Envar($name:state) no;
	}
    }
    if {[lsearch $Envar($name:allow) value] != -1} {
	set Envar($name:state) value;
    }
    if {$Envar($name:state) == "unknown"} {
	puts "$pfx: '$Envar($name:value)' seems inappropriate for '$Envar($name:name)'";
    }
}

################################################################################
# op_elookup
#
# Return the value of the environment variable option (name)
#
proc op_elookup {name} {

    global Envar;

    foreach tag $Envar(all) {			# search all
	if {$Envar($tag:name) == $name} {	# names match
	    switch $Envar($tag:state) {
		set -
		yes {
		    return "YES";
		}
		no {
		    return "NO";
		}
		value {
		    return $Envar($tag:value);
		}
		unset {
		    return "UNSET";		# erk, magic value
		}
	    }
	}
    }
    return "UNSET";				# erk, magic value
}

################################################################################
# op_init
#
# Read the ~/.pib file, set global options according to it and the
# environment.
#
proc op_init {} {

    global Options Envar env;

    # Start out by defining all the enviroment options we grok
    #
    #       Tag		Name			Allowed states		Default state	Default value
    op_edef usa		"USA_RESIDENT"		{yesno}			no;
    op_edef dir		"PORTSDIR"		{unset value}		value		"/usr/ports";
    op_edef ddir	"DISTDIR"		{unset value}		value		"/usr/ports/distfiles";
    op_edef msb		"MASTER_SITE_BACKUP"	{unset value} 		unset;
    op_edef msf		"MASTER_SITE_FREEBSD"	{set unset}		unset;
    op_edef fcmd	"FETCH_CMD"		{unset value}		unset;

    op_edef pdebug	"PATCH_DEBUG"		{set unset}		unset;
    op_edef motif	"HAVE_MOTIF"		{set unset}		unset;
    op_edef smotif	"MOTIF_STATIC"		{set unset}		unset;
    op_edef batch	"BATCH"			{set unset}		unset;
    op_edef interact	"INTERACTIVE"		{set unset}		unset;

    op_edef prefix	"PREFIX"		{unset value}		unset;
    op_edef mprefix	"MANPREFIX"		{unset value}		unset;
    op_edef mcomp	"NOMANCOMPRESS"		{set unset}		unset;
    op_edef noman	"NO_INSTALL_MANPAGES"	{set unset}		unset;
    op_edef pack	"PACKAGES"		{unset value}		unset;
    op_edef freg	"FORCE_PKG_REGISTER"	{set unset}		unset;

    op_edef ncld	"NOCLEANDEPENDS"	{set unset}		unset;

    # Set program options
    set Options(version_string) "\# PIB options v1";
    set Options(warn_no_md5) 1;
    set Options(warn_no_fdir) 0;
    set Options(warn_no_port) 0;
    set Options(warn_dup_names) 1;
    set Options(warn_del_errs) 1;

    set Options(do_deletions) 0;

    # if the user has a ~/.pib file...
    if {![catch {set fh [open "~/.pib" r]} ]} {

	gets $fh line;
	if {$line != $Options(version_string)} {
	    puts "Can't read ~/.pib file, incompatible format '$line'";
	} else {

	    while {[gets $fh line] >= 0} {		# read lines
		if {[string index $line 0] == "\#"} {	# ignore comments
		    continue ;
		}
		set tag [lindex $line 0];			# get first word

		# are we setting an environment variable?
		if {$tag == "env"} {
		    set value "";				# forget previous value
		    if {[scan $line "env %s %s %s" name state value] < 2} {
			puts "~/.pib: bad env spec '$line'";
			continue ;
		    }
		    set Envar($name:state) $state;
		    set Envar($name:value) $value;
		    continue ;
		}
		
		# is it a local option?
		if {[info exist Options($tag)]} {		# matches an option?
		    set Options($tag) [lrange $line 1 end];
		    continue ;
		}
		puts "~/.pib: bad line '$line'";
	    }
	}
	close $fh;
    }

    # scan the environment for overrides for environment variables
    foreach ev [array names env] {
	foreach name $Envar(all) {			# scan all we know
	    if {$Envar($name:name) == $ev} {		# env var matches one we know
		set Envar($name:value) $env($ev);	# get value
		op_estatus $name "environment";		# update status
	    }
	}
    }
    # and alow /etc/make.conf to override them again
    if {![catch {set fh [open "/etc/make.conf" r]}]} {

	while {[gets $fh line] >= 0} {		# read lines
	    if {[string index $line 0] == "\#"} {	# ignore comments
		continue ;
	    }
	    # look for TAG= VALUE
	    if {[scan $line "%\[^=\]=%s" tag val] == 2} {
		set tag [string trim $tag];
		set val [string trim $val];

		foreach name $Envar(all) {		# scan all we know
		    if {$Envar($name:name) == $tag} {	# tag matches one we know
			set Envar($name:value) $val;	# get value
			op_estatus $name "make.conf";	# update status
		    }
		}
	    }
	}
	close $fh;
    }
}

################################################################################
# op_save
#
# Writes the ~/.pib file
#
proc op_save {} {

    global Options Envar;

    if {[catch {set fh [open "~/.pib" w]} msg]} {
	tk_dialog .err "ERROR" $msg error 0 "Dismiss";
	return ;
    }

    puts $fh $Options(version_string);

    # internal options
    foreach var [lsort [array names Options]] {
	if {$var == "version_string"} {
	    continue ;				# makes no sense to emit this one
	}
	puts $fh "$var $Options($var)";
    }

    # environment variables
    foreach tag $Envar(all) {
	puts $fh "env $Envar($tag:name) $Envar($tag:state) $Envar($tag:value)";
    }

    close $fh;
}

################################################################################
# Toplevel browser
################################################################################
# br_selset
#
# Given the current browser path, load the selector 
#
proc br_selset {} {

    global Browser Ports Paths Options;

    # work out our current directory
    set cdir [path_getdir $Browser(cpath)];

    if {$cdir == $Browser(opath)} {		# got this up already?
	return ;
    }

    # This can take a little while, show we're busy
    set ocursor [$Browser(sellist) cget -cursor];
    $Browser(sellist) configure -cursor watch;
    update idletasks;

    set Browser(opath) $cdir;			# save new path

    # first nuke everything that's in it 
    $Browser(sellist) delete 0 end;

    # if the current is not the upward limit, add a '..' entry for going up with
    if {$Browser(cpath) != "[op_elookup PORTSDIR]/"} {
	$Browser(sellist) insert end "..";
    }

    # make a list of any paths below the current location
    set subdirs [array names Paths $cdir\*];

    # add these to the list
    set done_p "";
    foreach dir [lsort $subdirs] {
	set ncomp [next_pcomp $cdir $dir];
	if {[lsearch -exact $done_p $ncomp] == -1} {
	    $Browser(sellist) insert end $ncomp;
	    lappend done_p $ncomp;
	}
    }

    # if there are ports at this point in the tree, add them
    if {[info exists Paths($cdir)]} {

	foreach port [lsort $Paths($cdir)] {
	    $Browser(sellist) insert end $port;
	}
    }
    # restore the cursor
    $Browser(sellist) configure -cursor $ocursor;
}

################################################################################
# br_forward
#
# Go to the object selected in the selector.  If the current path is not a
# directory, pop the name off the end first.
#
# If the resulting path references a port, display it, otherwise go to the
# new directory.
#
proc br_forward {} {

    global Browser Paths;

    set index [$Browser(sellist) curselection];
    if {$index == ""} {				# nothing selected
	return ;
    }
    set newcomp [$Browser(sellist) get $index];

    if {$newcomp == ".."} {			# explict go-up command
	set Browser(history) "";
	br_back;				# go back instead
	return ;
    }

    # Work out new path
    set cdir [path_getdir $Browser(cpath)];
    append cdir $newcomp;

    set Browser(cpath) $cdir;
    if {[path_isdir $cdir]} {			# a directory?
	br_setport "";				# clear the port info
    } else {
	set pname $Paths($cdir);
	br_setport $pname;			# new port to detail
    }
    br_selset;					# reset the selector
    set Browser(history) "";			# forget any xref-following history
}

################################################################################
# br_back
#
# Go back up the xref history, if possible, or up a directory if not.
#
proc br_back {} {

    global Browser Ports Options;

    if {[llength $Browser(history)] > 0} {	# xref history?
	set newport [lindex $Browser(history) 0];
	set Browser(history) [lrange $Browser(history) 1 end];
	set Browser(cpath) $Ports($newport:path);	# go to new path
	br_setport $newport;				# show new port
	br_selset;					# new selection list
	return ;
    }

    set cdir [path_getdir $Browser(cpath)];
    if {$cdir == "[op_elookup PORTSDIR]/"} {	# can't go back before there
	return ;
    }

    set slen [expr [string length $cdir] -2];
    set cdir [string range $cdir 0 $slen];	# strip trailing slash
    set Browser(cpath) [path_getdir $cdir];	# back up again
    br_setport "";				# always clear
    br_selset;					# repopulate
}

################################################################################
# br_xref
#
# Take the port name selected in (list) and snap to it.  Add our current
# port to the xref history so we can come back.
#
proc br_xref {list} {

    global Browser Ports Paths;

    set index [$list curselection];
    if {$index == ""} {				# nothing selected
	return ;
    }
    set newport [$list get $index];

    if {![info exists Ports($newport:path)]} {
	
	tk_dialog .message "" "Don't know port '$newport'" error 0 "Dismiss";
	return ;				# we don't know about it
    }

    # Save the current port so we can go back to it
    lappend Browser(history) $Paths($Browser(cpath));

    set Browser(cpath) $Ports($newport:path);	# go to new path
    br_setport $newport;			# show new port
    br_selset;					# new selection list
}    

################################################################################
# br_search
#
# Post a search dialog, and allow the user to search for ports by any
# field(s)
#
proc br_search {} {

    global Browser;

    set w .brsearch;

    if {[winfo exists $w]} {			# already exists
	wm deiconify $w;			# do our best to become visible
	raise $w;
	tkwait visibility $w;
	focus $w;
	return ;
    }

    catch {destroy $w};
    toplevel $w;
    wm title $w "Search ports INDEX";

    # search term entry
    frame $w.for
    pack $w.for -side top -fill x;
    label $w.for.l -text "Seach for : ";
    entry $w.for.e -textvariable Browser(search:term) -width 30;
    pack $w.for.l -side left;
    pack $w.for.e -side left -fill x;

    # fields to search in
    frame $w.what;
    pack $w.what -side top -fill x;
    checkbutton $w.what.name -text "port name" -variable Browser(search:name) -anchor w;
    set Browser(search:name) 1;
    checkbutton $w.what.path -text "port directory path" -variable Browser(search:path) -anchor w;
    set Browser(search:path) 0;
    checkbutton $w.what.comment -text "comment" -variable Browser(search:comment) -anchor w;
    set Browser(search:comment) 0;
    checkbutton $w.what.maint -text "maintainer" -variable Browser(search:maint) -anchor w;
    set Browser(search:maint) 0;
    checkbutton $w.what.cat -text "categories" -variable Browser(search:cat) -anchor w;
    set Browser(search:cat) 0;
    checkbutton $w.what.bdep -text "build-depend" -variable Browser(search:bdep) -anchor w;
    set Browser(search:bdep) 0;
    checkbutton $w.what.rdep -text "run-depends" -variable Browser(search:rdep) -anchor w;
    set Browser(search:rdep) 0;
    pack $w.what.name $w.what.path $w.what.comment $w.what.maint $w.what.cat \
	$w.what.bdep $w.what.rdep -side top -fill x;

    # results
    frame $w.res;
    pack $w.res -side top -fill x;
    set sf $w.res;

    listbox $sf.list -width 30 -height 5 -yscrollcommand "$sf.scroll set";
    scrollbar $sf.scroll -command "$sf.list yview";
    bind $sf.list <Double-1> "br_searchsel $sf.list";
    pack $sf.list -side left -fill both -expand y;
    pack $sf.scroll -side right -fill y;

    # buttons
    frame $w.butt;
    pack $w.butt -side bottom -fill x;
    button $w.butt.search -text "Search" -command "br_dosearch $w.res.list" -width 10;
    button $w.butt.hlp -text "Help" -command "help search";
    button $w.butt.dis -text "Dismiss" -command "destroy $w" -width 10;
    pack $w.butt.search -side left;
    pack $w.butt.dis $w.butt.hlp -side right;

    # try to be intuitive
    focus $w.for.e;
    bind $w.for.e <Key-Return> "br_dosearch $w.res.list";
}

################################################################################
# br_dosearch
#
# Search the INDEX for stuff.
#
proc br_dosearch {list} {

    global Browser Paths Ports;

    $list delete 0 end;					# delete list contents

    foreach path [array names Paths] {			# walk list of all ports
	
	set name $Paths($path);				# port name
	set match 0;					# not a match yet

	foreach what "name path comment maint cat bdep rdep" {
	    if {$Browser(search:$what)} {
		if {[ regexp -nocase ".\*$Browser(search:term).\*" $Ports($name:$what)]} {
		    set match 1;
		}
	    }
	}
	if {$match} {					# got a match
	    $list insert end $name;
	}
    }
}
    
################################################################################
# br_searchsel
#
# Given (list) containing port names, find which one has just been 
# selected and go there.
#
proc br_searchsel {list} {

    global Browser Ports Paths;

    set index [$list curselection];
    if {$index == ""} {				# nothing selected
	return ;
    }
    set newport [$list get $index];
    set newpath $Ports($newport:path);

    set Browser(cpath) $newpath;
    br_setport $newport;			# new port to detail
    br_selset;					# reset the selector
    set Browser(history) "";			# forget any xref-following history
}

################################################################################
# br_setport
#
# Given (name), display information about the port.  If (name) is an 
# empty string, display empties.
#
# We crosslink to the distfile browser here, and if it is in 'for port'
# mode, call df_update to update the information for the new port.
#
proc br_setport {name} {

    global Browser DBrowser Maker Ports;

    $Browser(pcomment) configure -state normal;	# write-enable
    # clean up
    set Browser(pname) "";
    $Browser(pcomment) delete 1.0 end;
    set Browser(maint) "";
    $Browser(bdlist) delete 0 end;
    $Browser(rdlist) delete 0 end;

    if {[array names Ports $name:\*] != ""} {	# port matching name?
	set Browser(pname) "$Ports($name:cat) : $name";
	set Browser(dpath) "$Ports($name:dpath)";
	$Browser(pcomment) insert end $Ports($name:comment);
	set Browser(maint) $Ports($name:maint);
	foreach bd $Ports($name:bdep) {
	    $Browser(bdlist) insert end $bd;
	}
	foreach rd $Ports($name:rdep) {
	    $Browser(rdlist) insert end $rd;
	}
	br_desc "old";		# if we can, browse the description
	# Update the Maker's status
	set Maker(port) $name;
    }
    $Browser(pcomment) configure -state disabled;	# write-protect

    br_portstat $name;		# update the status indicators

    # Check the distfile browser state
    if {[info exists DBrowser(mode:for)]} {
	if {$DBrowser(mode:for)} {
	    df_update;
	}
    }
}

################################################################################
# br_desc
#
# Display the +DESCR file for the currently selected path.  If (mode) is "new",
# create the reader if it doesn't exist.
#
proc br_desc {{mode old}} {

    global Browser;

    if {[path_isdir $Browser(cpath)] || ($Browser(cpath) == "")} {
	return ;				# not a port
    }
    
    set w ".descreader";

    set fname "$Browser(dpath)";
    if {![file readable $fname]} {
	if {$mode == "new"} {
	    tk_dialog .message "" "Can't open the file :\n$fname\nto read port description." \
		error 0 "Dismiss";
	}
	catch {destroy $w};			# throw it away if possible
	return ;
    }

    # reader window
    if {![winfo exists $w]} {			# need to create a new one
	if {$mode != "new"} {
	    return ;				# no reader, not creating a new one
	}
	catch {destroy $w};
	toplevel $w;

	# text & scrollbar
	frame $w.t;
	pack $w.t -side top -fill both -expand y;
	text $w.t.text -width 80 -height 20 -wrap none \
	    -yscrollcommand "$w.t.scroll set";
	scrollbar $w.t.scroll -command "$w.t.text yview";
	pack $w.t.text -side left -fill both -expand y;
	pack $w.t.scroll -side right -fill y;

	# dismiss-button
	frame $w.b
	pack $w.b -side bottom -fill x;
	button $w.b.dis -text "Dismiss" -width 10 -command "destroy $w";
	pack $w.b.dis -side right;    
    }

    wm title $w $Browser(cpath);
    # clean it out
    $w.t.text configure -state normal;
    $w.t.text delete 1.0 end;

    # read file in
    set fh [open $fname "r"];
    while {[gets $fh line] >= 0} {
	$w.t.text insert end "$line\n"
    }
    close $fh
    $w.t.text configure -state disabled;
}
    
################################################################################
# br_portstat
#
# Update the status indicators for (name)
#
proc br_portstat {name} {

    global Browser Ports;

    set on "";				# none on yet

    if {$name != ""} {			# actually references a port?
	set path $Ports($name:path);	# where the port lives

	if {![file isdirectory $path]} {	# port present?
	    lappend on "stat:absent";		# not here
	} else {
	    if {![file isdirectory "$path/work"]} {	# work directory present?
		lappend on "stat:clean";		# all tidy
	    } else {
		foreach op [list "extract" "patch" "configure" "build" "install" "package"] {
		    set indicator [format "%s/work/.%s_done" $path $op];
		    if {[file exists $indicator]} {
			lappend on "stat:$op";
		    }
		}
	    }
	}

	# look for registration in /var/db/pkg as well.  We have to be rude
	# to do this because of bizarre permissions there.
	set patt "/var/db/pkg/$name";
	if {[glob -nocomplain $patt\*] == $patt} {
	    lappend on "stat:install";
	}
    }
    # Set all controls
    foreach op [array names Browser stat:*] {
	if {[lsearch -exact $on $op] != -1} {
	    $Browser($op) configure -foreground "dark green";
	} else {
	    $Browser($op) configure -foreground gray;
	}
    }
}


################################################################################
# path_isdir
#
# Returns true if (path) is a directory, rather than a port path.  The
# path always has a trailing slash if it does not reference a port.
#
proc path_isdir {path} {

    if {[lastchar $path] == "/"} {
	return 1;
    }
    return 0;
}

################################################################################
# path_getdir
#
# returns (path) with everything after the last slash removed
#
proc path_getdir {path} {

    set lspos [string last "/" $path];
    return [string range $path 0 $lspos];
}

################################################################################
# next_pcomp
#
# Given a partial path (path) and a full path (full), return the next 
# component of the full path.
#
proc next_pcomp {path full} {

    # get prefix length
    set spos [string length $path];

    # discard prefix
    set tail [string range $full $spos end];

    set epos [string first "/" $tail];
    if {$epos == -1} {				# no seperator
	set epos end;
    }
    return [string range $tail 0 $epos];
}

################################################################################
# Maker
################################################################################
# mk_create
#
# Creates the Maker window.
#
proc mk_create {} {

    global Maker Browser;

    set w ".mk";
    
    if {[winfo exists $w]} {			# already exists
	return ;
    }

    catch {destroy $w};
    toplevel $w;
    wm title $w "Port Maker";

    # frame across the top with some info
    frame $w.head;
    pack $w.head -side top -fill x;
    
    entry $w.head.p -textvariable Maker(port) -relief raised -bd 1 \
	-width 20 -state disabled;
    entry $w.head.s -textvariable Maker(status) -relief raised -bd 1 \
	-state disabled;
    pack $w.head.p -side right;
    pack $w.head.s -side left -fill x -expand y;
    set Maker(status) "Idle";
    if {![info exist Maker(port)]} {
	set Maker(port) "";			# no port currently selected
    }

    frame $w.thalf;
    pack $w.thalf -side top -fill x;
    
    # Crank up an xterm and steal it into our frame
    tkSteal $w.thalf.xterm -relief sunken -command \
	[list xterm -w 0 -font fixed -xrm "xterm*allowSendEvents:true" -geometry 100x50 \
	     -sb -sl 2000 -n makerslave -T makerslave -e /bin/sh -I] \
	-name makerslave -width 80 -height 25;
    pack $w.thalf.xterm -side left -fill both -expand y;
    set Maker(xterm) $w.thalf.xterm;

    # Pack a frame with some command buttons
    frame $w.thalf.cmd;
    pack $w.thalf.cmd -side right -fill y;
    set sf $w.thalf.cmd;

#    button $sf.remove    -text "Remove"    -anchor w -command mk_remove;
    button $sf.clean     -text "Clean"     -anchor w -command "mk_do clean";
    button $sf.fetch     -text "Fetch"     -anchor w -command "mk_do fetch";
    button $sf.extract   -text "Extract"   -anchor w -command "mk_do extract";
    button $sf.patch     -text "Patch"     -anchor w -command "mk_do patch";
    button $sf.configure -text "Configure" -anchor w -command "mk_do configure";
    button $sf.build     -text "Build"     -anchor w -command "mk_do build";
    button $sf.install   -text "Install"   -anchor w -command "mk_do install";
    button $sf.package   -text "Package"   -anchor w -command "mk_do package";

    pack $sf.clean $sf.fetch $sf.extract $sf.patch $sf.configure \
	$sf.build $sf.install $sf.package -side top -fill x;
    set Maker(cmds) $sf;

    # frame at the bottom for more buttons
    frame $w.butt;
    pack $w.butt -side bottom -fill x;

    button $w.butt.accept -text "Accept" -width 10 -state disabled \
	-command mk_accept;
    pack $w.butt.accept -side left;
    set Maker(accept) $w.butt.accept;

    button $w.butt.dis -text "Dismiss" -width 10 \
	-command "$w.thalf.xterm stoptksteal; destroy $w";
    button $w.butt.help -text "Help" -command "help maker";
    pack $w.butt.dis $w.butt.help -side right;
}

################################################################################
# mk_do
#
# Perform (func) on the current port, if one exists
#
proc mk_do {func} {

    global Maker Ports Envar;

    if {$Maker(port) == ""} {
	return ;				# nothing to do
    }

    # disable other command buttons; don't want to confuse things.
    mk_mode active;

    set pname $Maker(port);
    mk_status "Setup for 'make $func' on '$pname'";

    # chdir to desired path
    mk_command "cd $Ports($pname:path)";	# run this one

    # setup the environment
    set envcmd "";
    foreach tag $Envar(all) {
	set name $Envar($tag:name);
	switch $Envar($tag:state) {
	    set -
	    yes {
		append envcmd "export $name=YES; ";
	    }
	    no {
		append envcmd "export $name=NO; ";
	    }
	    value {
		append envcmd "export $name=$Envar($tag:value); ";
	    }
	    unset {
		append envcmd "unset $name; ";
	    }
	}
    }
    mk_command $envcmd;				# set the env (messy)

    # Run the command
    mk_status "Run 'make $func' on '$pname'";
    mk_command "make $func";
}

################################################################################
# mk_accept
#
# User has accepted results of command, reenable command buttons
#
proc mk_accept {} {
    mk_mode inactive;
}

################################################################################
# mk_status
#
# Announce (msg) and make it our current status
#
proc mk_status {msg} {

    global Maker;

    mk_command "\#\#\#\#\# $msg \#\#\#\#\#"
    set Maker(status) $msg;
    update;
}

################################################################################
# mk_command
#
# Send (cmd) to our slave xterm.
#
proc mk_command {cmd} {

    global Maker;

    $Maker(xterm) sendstring "$cmd\r";
}

################################################################################
# mk_mode
#
# Go to (mode) mode; inhibit all the command buttons and enable the
# "accept" buttons or vice-versa.
#
proc mk_mode {mode} {

    global Maker;

    switch $mode {
	active {
	    set cmode "disabled";
	    set amode "normal";
	}
	inactive {
	    set cmode "normal";
	    set amode "disabled";
	}
	default {
	    # oops
	    return ;
	}
    }
    foreach w [winfo children $Maker(cmds)] {
	if {[winfo class $w] == "Button"} {
	    $w configure -state $cmode;
	}
    }
    $Maker(accept) configure -state $amode;
}


################################################################################
# INDEX parser
################################################################################
# read_index
#
# Parse the INDEX-format file on (fh), build the Ports() and Paths() arrays.
#
proc read_index {fh pwin} {

    global Ports Paths Status Options;

    set Status(nports) 0;

    set errs "";
    while {[gets $fh line] >= 0} {

	set elems [split $line "|"];

	set pname [lindex $elems 0];		# get port name

	if {[info exists Ports($pname:name)]} {	# oops, duplicate
	    if {$Options(warn_dup_names)} {
		lappend errs "$pname: duplicate port name.";
	    }
	    continue ;				# ignore it (bad idea?)
	}

	# Populate arrays
	set Ports($pname:name)	  $pname;
	set Ports($pname:path)    [lindex $elems 1];
	set Ports($pname:prefix)  [lindex $elems 2];
	set Ports($pname:comment) [lindex $elems 3];
	set Ports($pname:dpath)   [lindex $elems 4];
	set Ports($pname:maint)   [lindex $elems 5];
	set Ports($pname:cat)  	  [lindex $elems 6];
	set Ports($pname:bdep)    [lindex $elems 7];
	set Ports($pname:rdep)    [lindex $elems 8];
	set Ports($pname:distfiles) "";			# don't know which distfiles it requires yet
	set Paths($Ports($pname:path)) $pname;		# xref path to port name

#	puts "$Ports($pname:path) -> ($Ports($pname:cat))$pname";

	# Update the progress counter
	incr Status(nports);
	update_pwin $pwin $Status(nports);
    }
    if {$errs != ""} {
	create_listwin "Errors while processing INDEX" df_idx_errs [lsort $errs];
    }
}

################################################################################
# Distfile functions
################################################################################
# df_browser
#
# Generates the distfile browser widget.
#
proc df_browser {} {

    global DBrowser Status;

    set w ".df";
    
    if {[winfo exists $w]} {			# already exists
	return ;
    }

    catch {destroy $w};
    toplevel $w;
    wm title $w "Ports Distfile browser";

    # command/control frame
    frame $w.comm;
    pack $w.comm -side top -fill x;

    # Controls
    set sf $w.comm;
    checkbutton $sf.curr -text "Current" -variable DBrowser(mode:curr) \
	-command df_update -onvalue "current" -offvalue "no" \
	-activeforeground white -activebackground black;
    checkbutton $sf.stale -text "Stale" -variable DBrowser(mode:stale) \
	-foreground darkgreen -command df_update -onvalue "stale" -offvalue "no" \
	-activeforeground white -activebackground darkgreen;
    checkbutton $sf.missing -text "Missing" -variable DBrowser(mode:missing) \
	-foreground darkblue -command df_update -onvalue "missing" -offvalue "no" \
	-activeforeground white -activebackground darkblue;
    checkbutton $sf.for -text "For selected port" -variable DBrowser(mode:for) \
	-command df_update;
    button $sf.dump -text "Dump" -command "df_dump distfiles.dump" -width 10;
    button $sf.re -text "Rescan" -command "df_gather; df_update" -width 10;

    # default is all on, for all files
    set DBrowser(mode:curr) "current";
    set DBrowser(mode:stale) "stale";
    set DBrowser(mode:missing) "missing";
    set DBrowser(mode:for) 0;

    pack $sf.curr $sf.stale $sf.missing $sf.for -side left;
    pack $sf.re $sf.dump -side right;

    # Distfile list area
    frame $w.list;
    pack $w.list -side top -fill both -expand y;
    set sf $w.list;

    text $sf.flist -width 40 -height 15 -yscrollcommand "$sf.scroll set" \
	-state disabled -wrap none;
    $sf.flist tag configure "stale" -foreground darkgreen;
    $sf.flist tag configure "missing" -foreground darkblue;

    text $sf.plist -width 50 -height 15 -yscrollcommand "$sf.scroll set" \
	-state disabled -wrap none;
    $sf.plist tag configure "reason" -foreground red;
    # cosmetics
    $sf.plist configure -cursor arrow;
    # toggle individual tags
    bind $sf.plist <1> \
	[format \
	     "%%W tag remove sel 0.0 end; if {\[df_tagtoggle %s %s %%x %%y\]} { break ; }" \
	     $sf.flist $sf.plist \
	    ];

    # Bind scrollbar to both text widgets with a helper
    scrollbar $sf.scroll -command "df_br_lset $sf.flist $sf.plist";
    set DBrowser(flist) $sf.flist;
    set DBrowser(plist) $sf.plist;

    pack $sf.flist -side left -fill y;
    pack $sf.plist -side left -fill both -expand y;
    pack $sf.scroll -side right -fill y;

    # Function buttons down the bottom
    frame $w.butt;
    pack $w.butt -side bottom -fill x;
    button $w.butt.del -text "Delete" -command "df_delstale $sf.flist";
    button $w.butt.ta -text "Tag" -command "df_tag $sf.flist 1";
    button $w.butt.uta -text "Untag" -command "df_tag $sf.flist 0";
    pack $w.butt.del $w.butt.ta $w.butt.uta -side left;

    button $w.butt.hlp -text "Help" -command "help distfiles";
    button $w.butt.dis -text "Dismiss" -width 10 -command "destroy $w";
    pack $w.butt.dis $w.butt.hlp -side right;

    # Collect information and then update
    if {$Status(df:info) == "invalid"} {
	set ocursor [$w cget -cursor];
	$w configure -cursor watch;
	update idletasks;
	df_gather;
	$w configure -cursor $ocursor;
    }
    df_update;
}

################################################################################
# df_br_lset
#
# Helper function to set y position on two objects
#
proc df_br_lset {w1 w2 args} {
    eval $w1 yview $args;
    eval $w2 yview $args;
}

################################################################################
# df_update
#
# Update the distfile browser.  Can be called safely if the browser is not
# up, and runs reasonably quickly.
#
proc df_update {} {

    global DBrowser Browser lastset AllDistfiles DistfileStatus Paths Ports;
    global Checksums TagList;

    set w .df;
    if {![winfo exists $w]} {
	return ;			# window gone?
    }

    set ocursor [$w cget -cursor];
    $w configure -cursor watch;
    update idletasks;
    
    # clean out the previous date
    $DBrowser(flist) configure -state normal;
    $DBrowser(plist) configure -state normal;
    $DBrowser(flist) delete 0.1 end;
    $DBrowser(plist) delete 0.1 end;

    # determine which distfiles might be relevant
    if {$DBrowser(mode:for)} {		# current port only?
	if {[path_isdir $Browser(cpath)]} {
	    $w configure -cursor $ocursor;
	    return ;			# not on a port
	}
	set pname $Paths($Browser(cpath));
	set fullist [lsort $Ports($pname:distfiles)];

    } else {				# all distfiles 
	set fullist [lsort [array names AllDistfiles]];
    }

    # Walk the list one at a time and decide if we want to see it.
    # This loop should be as tight as possible, as it has to
    # process the entire list of possible and present distfiles.
    set last "";
    foreach df $fullist {
	set state $DistfileStatus($df);

	if {($DBrowser(mode:curr) == $state) ||
	    ($DBrowser(mode:stale) == $state) ||
	    ($DBrowser(mode:missing) == $state)} {

	    $DBrowser(flist) insert end "$df\n" $state;

	    if {$state == "stale"} {
		if {[info exist TagList($df)]} {
		    $DBrowser(plist) insert end "\[x\] " tagmark;
		} else {
		    $DBrowser(plist) insert end "\[ \] " tagmark;
		}
		if {$Checksums($df) == ""} {
		    $DBrowser(plist) insert end "(orphan)" reason;
		} else {
		    $DBrowser(plist) insert end "(bad MD5)" reason;
		}
	    } else {
		$DBrowser(plist) insert end "    ";
	    }
	    catch {
		foreach port $AllDistfiles($df) {
		    $DBrowser(plist) insert end " $port";
		}
	    }
	    $DBrowser(plist) insert end "\n";
	}
    }
    # write protect again
    $DBrowser(flist) configure -state disabled;
    $DBrowser(plist) configure -state disabled;

    $w configure -cursor $ocursor;
}

################################################################################
# df_delstale
#
# Delete stale distfiles listed in the TagList, either for a single port 
# ('for' mode), or for all ports.  Distfile names are extracted from
# (list)
#
proc df_delstale {list} {

    global AllDistfiles Checksums DistfileStatus Options TagList;

    set errs "";
    if {!$Options(do_deletions)} {
	lappend errs "Deletions disabled, listing files that would be deleted.";
    }

    for {set line 1} {1} {incr line} {
	# calculate text indices
	set from "$line.0"
	set to "$line.end";

	set df [$list get $from $to];
	if {$df == ""} {
	    break ;					# end of the list
	}

	if {($DistfileStatus($df) != "stale") ||
	    (![info exist TagList($df)])} {
	    continue ;			# don't delete it
	}

	set fname "[op_elookup DISTDIR]/$df";
	if {$Options(do_deletions)} {
	    if {[catch {exec rm -f $fname} msg]} {
		if {$Options(warn_del_fail)} {
		    lappend errs $msg;
		}
	    } else {
		# Its gone; forget about it.
		unset AllDistfiles($df);		# must be here
		unset DistfileStatus($df);		# and here
		unset TagList($df);			# and here
		catch {unset Checksums($df)};		# may not be here
	    }
	} else {
	    lappend errs "delete '$fname'";
	}
    }
    # errors while we were processing?
    if {$errs != ""} {
	create_listwin "Errors while deleting stale distfiles" df_del_errs [lsort $errs];
	wait_listwin df_del_errs;
    }
    df_update;
}
    
################################################################################
# df_gather
#
# Walk the list of ports and parse the md5 files, build the AllDistfiles and
# Checksums arrays.  Also walk the collection of present distfiles.
#
proc df_gather {} {

    global AllDistfiles DistfileStatus Checksums Paths Ports Status Options;

    set Status(df:info) "invalid";

    # Delete any previous state
    catch {unset AllDistfiles};
    catch {unset Checksums};

    set errs "";
    set pports 0;
    set psize 0;
    set stamp [clock seconds];
    create_pwin "df_getdinfo" "Reading distfile information: %s";
    set Status(adfiles) 0;
    set distdir [op_elookup DISTDIR];

    # process all ports in the INDEX
    foreach path [array names Paths] {
	# get port's real name
	set pname $Paths($path);

	# update the progress indicator
	incr pports;
	set rate [expr int($psize / (1000 * ([clock seconds] - $stamp) + 1))];
	update_pwin df_getdinfo [format "%d/%d ports, %dK/sec" $pports $Status(nports) $rate];

	# Find the md5 file.
	if {[catch {set fh [open "$path/distinfo" "r"]}]} {

	    # Can't find an distinfo file; give up
	    if {[file isdirectory "$path"]} {
		if {$Options(warn_no_md5)} {
		    lappend errs "$pname: '$path/distinfo' not found";
		}
	    } else {
		if {$Options(warn_no_port)} {
		    lappend errs "$pname: directory missing.";
		}
	    }
	    continue ;				# if no port info, assume it's not wanted
	}

	# initialise state for this port
	set Ports($pname:distfiles) "";		# list of distfiles for the port

	# Process md5 file to get distfiles and checksums
	while {[gets $fh line] >= 0} {

	    if {[scan $line "MD5 (%\[^)\]) = %s" df ck] == 2} {	# get name and checksum
		if {[info exists Checksums($df)]} {		# already have a checksum?
		    if {$Checksums($df) != $ck} {
			lappend errs "$pname: checksum for '$df' does not match previous '$Checksums($df)'";
		    }
		}
		set Checksums($df) $ck;				# remember checksum for later
		lappend AllDistfiles($df) $pname;		# remember port uses this distfile
		if {[catch {set DistfileStatus($df) [df_checkchecksum $df "$distdir/$df"]} msg]} {
		    lappend errs "$pname: error checking checksum '$msg'";
		    set DistfileStatus($df) "missing";		# XXX ?
		} 
		if {$DistfileStatus($df) != "missing"} {
		    # file has been scanned, add size to our records
		    incr psize [file size "$distdir/$df"];
		}
		lappend Ports($pname:distfiles) $df;		# add to list for this port
		incr Status(adfiles);
	    }
	}
	close $fh;
    }
    destroy_pwin df_getdinfo;
#    lappend errs [format "%dK in %d seconds" [expr $psize/1000] [expr [clock seconds] - $stamp]];

    # process all existing distfiles; if they're not already known then 
    # they must be stale
    set ddlen [expr [string length $distdir] + 1];
    if {[catch {set rawlist [df_getfiles_dir $distdir]} msg]} {
	lappend errs "Error scanning for stale distfiles: $msg";
    } else {
	# look to see if we know this file
	foreach rdf $rawlist {
	    # XXX strip off DISTDIR
	    set df [string range $rdf $ddlen end];
	    if {![info exists AllDistfiles($df)]} {		# no, it's stale
		set AllDistfiles($df) "";
		set DistfileStatus($df) "stale";
		set Checksums($df) "";
	    }
	}
    }
    # errors while we were processing?
    if {$errs != ""} {
	create_listwin "Errors while processing ports distfile information" df_pd_errs [lsort $errs];
	wait_listwin df_pd_errs;
    }
    set Status(df:info) "valid";

}

################################################################################
# df_getfiles_dir
#
# Recursive routine to produce list of files rooted at (dir)
#
proc df_getfiles_dir {dir} {

    set files "";

    set list [glob -nocomplain "$dir/*"];

    foreach ent $list {

	if {[file isdirectory $ent]} {		# subdirectory?
	    set files [concat $files [df_getfiles_dir $ent]];
	}

	if {[file isfile $ent]} {
	    lappend files $ent;
	}
    }
    return $files;
}

################################################################################
# df_checkchecksum
#
# Verify that the checksum for the distfile (df) matches the value claimed
# in Checksums, returns "missing", "stale" or "current" as appropriate.
#
# (path) is the full path to (df)
#
proc df_checkchecksum {df path} {

    global Checksums;

    if {![file readable $path]} {
	return "missing";
    }
    if {$Checksums($df) != "IGNORE"} {		# do we care?
	if {[catch {set result [exec /sbin/md5 $path]}]} {
	    error "MD5 failed";
	}
	if {[set rck [lindex $result 3]] == "" } {
	    error "can't parse MD5 output";		# bad md5 output
	}
	if {$Checksums($df) != $rck} {
	    return "stale";
	}
    }
    return "current";
}

################################################################################
# df_dump
#
# Write a dump listing of the contents of the distfile browser to (fname).
#
#
proc df_dump {fname} {

    global DBrowser;

    if {![winfo exists $DBrowser(plist)]} {
	return ;			# lists gone
    }

    if {[file isdirectory $fname]} {
	set fname [format "%s/distfiles.lst" $fname];
    }
    if {[catch {set fh [open $fname "w"]} msg]} {
    	tk_dialog .err "ERROR" $msg error 0 "Dismiss";
	return ;
    }
    puts $fh [format "%-35s %s" "Distfile" "Ports"];
    puts $fh [format "%-35s %s" "========" "====="];

    for {set line 1} {1} {incr line} {
	# calculate text indices
	set from "$line.0"
	set to "$line.end";

	set df [$DBrowser(flist) get $from $to];
	set pl [$DBrowser(plist) get $from $to];
	if {$df == ""} {			# hit the end?
	    break ;
	}
	puts $fh [format "%-35s %s" [string trim $df] [string trim $pl]];
    }
    close $fh;
}

################################################################################
# df_tagtoggle
#
# Toggle the tag state for the distfile referenced by (x) (y) in (dwin)/(pwin).
#
proc df_tagtoggle {dwin pwin x y} {

    global TagList;

    set index [$pwin index @$x,$y];		# convert x,y to line.char
    scan $index "%d.%d" line col;	       	# get line
    if {$col > 2} {
	return 0;				# outside boxes, go for 'normal'
    }

    set df [$dwin get $line.0 $line.end];	# get distfile
    if {$df == ""} {
	return 0;				# oops!
    }
    if {[$pwin get $line.0] != "\["} {		# not switchable
	return 0;
    }
    set state [$pwin get $line.1];		# get current tag state
    if {$state == " "} {			# not tagged
	set state "x";
	set TagList($df) "yes";
    } else {					# tagged or broken
	set state " ";
	if {[info exist TagList($df)]} {	# was tagged?
	    unset TagList($df);
	}
    }
    $pwin configure -state normal;		# unlock
    $pwin delete $line.1;			# frob
    $pwin insert $line.1 $state tagmark;
    $pwin configure -state disabled;		# relock
    return 1;
}

################################################################################
# df_tag
#
# tag/untag all the distfiles listed in (list)
#
proc df_tag {list {mode 1}} {

    global TagList;

    for {set line 1} {1} {incr line} {
	# calculate text indices
	set from "$line.0"
	set to "$line.end";

	set df [$list get $from $to];
	if {$df == ""} {
	    break ;
	}
	if {$mode} {				# tag or untag?
	    set TagList($df) "yes";
	} else {
	    catch {unset TagList($df)}
	}
    }
    df_update;
}

################################################################################
# Help
################################################################################
# help
#
# If we know help for (what), put it up in a window
#
proc help {what} {

    global help;

    if {[info exists help($what)]} {
	create_listwin "Help" help $help($what);
    }
}

################################################################################
# help_init
#
# Sets the help for (what) to (text).
# (text) should be a list, with one list element per line.
#
proc help_init {what text} {

    global help;

    set help($what) $text;
}

################################################################################
# Utility functions
################################################################################
# create_pwin
#
# Create and realise a small progess-indicator window
#
proc create_pwin {tag fmt {ivalue ""}} {

    global progwin;

    # toast it just in case
    catch {wm destroy $progwin($tag:path)};

    # Name and create a toplevel
    set w .pwin_$tag;
    toplevel $w;
    wm title $w "Progress...";

    # Establish state
    set progwin($tag:path) $w;
    set progwin($tag:value) $ivalue;
    set progwin($tag:fmt) $fmt;
    set progwin($tag:text) [format $progwin($tag:fmt) $progwin($tag:value)];

    # Label for message
    label $w.msg -textvariable progwin($tag:text);
    pack $w.msg -side top;

    # Make it visible
    w_popcentre $w;

    # schedule updates
    after 500 "autoupdate_pwin $tag";
    
}

################################################################################
# destroy_pwin
#
# Remove the progress window for (tag)
#
proc destroy_pwin {tag} {

    global progwin;

    catch {destroy $progwin($tag:path)};
}

################################################################################
# autoupdate_pwin
#
# Update the progress information for (tag), allow update events
#
proc autoupdate_pwin {tag} {

    global progwin;

    if {[winfo exists $progwin($tag:path)]} {	# still there?

	# update text
	set progwin($tag:text) [format $progwin($tag:fmt) $progwin($tag:value)];
	
	# run again sometime
	after 500 "autoupdate_pwin $tag";
    }
}
	
################################################################################
# update_pwin
#
# Update the value for progress window (tag) with (val)
#
proc update_pwin {tag val} {

    global progwin;

    set progwin($tag:value) $val;

    # allow the background updates to happen
    update;
}

################################################################################
# create_listwin
#
# Create (if required) a window identified by (tag) with a scrolling 
# region containing the entities in (list), one per line.  Use (title) as the
# window title.
#
proc create_listwin {title tag list} {

    set w .lwin_$tag;
    
    if {![winfo exists $w]} {			# need to create a new one
	catch {destroy $w};
	toplevel $w;
	wm title $w $title;

	# text & scrollbar
	frame $w.t;
	pack $w.t -side top -fill both -expand y;
	text $w.t.text -width 80 -height 20 -wrap none \
	    -yscrollcommand "$w.t.scroll set";
	scrollbar $w.t.scroll -command "$w.t.text yview";
	pack $w.t.text -side left -fill both -expand y;
	pack $w.t.scroll -side right -fill y;

	# dismiss-button
	frame $w.b
	pack $w.b -side bottom -fill x;
	button $w.b.dis -text "Dismiss" -width 10 -command "destroy $w";
	pack $w.b.dis -side right;    
    }

    # clean it out
    $w.t.text configure -state normal;
    $w.t.text delete 1.0 end;

    foreach ent $list {
	$w.t.text insert end "$ent\n";
    }
    $w.t.text configure -state disabled;
    return $w.t.text;
}

################################################################################
# wait_listwin
#
# Wait for the list window associated with (tag) to go away
#
proc wait_listwin {tag} {

    set w ".lwin_$tag";
    if {[winfo exists $w]} {

	tkwait window $w;
    }
}

################################################################################
# w_popcentre
#
# Slap (w) in the middle of the screen, on top of everything.  Stolen
# (originally) from some of the Tk library code.
#
proc w_popcentre {w} {

    # Withdraw the window, then update all the geometry information
    # so we know how big it wants to be, then center the window in the
    # display and de-iconify it.

    wm withdraw $w
    update idletasks
    set x [expr [winfo screenwidth $w]/2 - [winfo reqwidth $w]/2 \
            - [winfo vrootx [winfo parent $w]]]
    set y [expr [winfo screenheight $w]/2 - [winfo reqheight $w]/2 \
            - [winfo vrooty [winfo parent $w]]]
    wm geom $w +$x+$y
    wm deiconify $w
}

################################################################################
# lastchar
#
# returns the last character of (str)
#
proc lastchar {str} {

    set p [expr [string length $str] -1];
    return [string index $str $p];
}


################################################################################
main;
################################################################################
